@@ -1,5 +1,6 @@
 /* Man page to help file converter
-   Copyright (C) 1994, 1995 Janne Kukonlehto
+   Copyright (C) 1994, 1995, 1998, 2000, 2001, 2002, 2003, 2004, 2005,
+   2007 Free Software Foundation, Inc.
 		 2002  Andrew V. Samoilov
 		 2002  Pavel Roskin
 
@@ -15,15 +16,21 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
+
+/** \file man2hlp.c
+ *  \brief Source: man page to help file converter
+ */
 
 #include <config.h>
+
+#include <stdarg.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include <stdarg.h>
 #include <string.h>
 
 #include <glib.h>
+
 #include "help.h"
 
 #define BUFFER_SIZE 256
@@ -180,7 +187,7 @@ static int
 string_len (const char *buffer)
 {
     static int anchor_flag = 0;	/* Flag: Inside hypertext anchor name */
-    static int link_flag = 0;	/* Flag: Inside hypertext link target name */
+    static int lc_link_flag = 0;	/* Flag: Inside hypertext link target name */
     int backslash_flag = 0;	/* Flag: Backslash quoting */
     int c;			/* Current character */
     int len = 0;		/* Result: the length of the string */
@@ -188,9 +195,9 @@ string_len (const char *buffer)
     while (*(buffer)) {
 	c = *buffer++;
 	if (c == CHAR_LINK_POINTER)
-	    link_flag = 1;	/* Link target name starts */
+	    lc_link_flag = 1;	/* Link target name starts */
 	else if (c == CHAR_LINK_END)
-	    link_flag = 0;	/* Link target name ends */
+	    lc_link_flag = 0;	/* Link target name ends */
 	else if (c == CHAR_NODE_END) {
 	    /* Node anchor name starts */
 	    anchor_flag = 1;
@@ -207,7 +214,7 @@ string_len (const char *buffer)
 	}
 	backslash_flag = 0;
 	/* Increase length if not inside anchor name or link target name */
-	if (!anchor_flag && !link_flag)
+	if (!anchor_flag && !lc_link_flag)
 	    len++;
 	if (anchor_flag && c == ']') {
 	    /* Node anchor name ends */
@@ -454,9 +461,8 @@ handle_alt_font (char *buffer)
 }
 
 /* Handle .IP and .TP commands.  is_tp is 1 for .TP, 0 for .IP */
-/* buffer is not used now */
 static void
-handle_tp_ip (char *buffer, int is_tp)
+handle_tp_ip (int is_tp)
 {
     if (col > 0)
 	newline ();
@@ -538,9 +544,9 @@ handle_command (char *buffer)
 	*w = 0;
 	print_string (buffer);
     } else if (strcmp (buffer, ".TP") == 0) {
-	handle_tp_ip (buffer, 1);
+	handle_tp_ip (1);
     } else if (strcmp (buffer, ".IP") == 0) {
-	handle_tp_ip (buffer, 0);
+	handle_tp_ip (0);
     } else if (strcmp (buffer, ".\\\"TOPICS") == 0) {
 	if (out_row > 1) {
 	    print_error
@@ -601,6 +607,7 @@ handle_link (char *buffer)
     static char old[80];
     int len;
     char *amp;
+    const char *amp_arg;
 
     switch (link_flag) {
     case 1:
@@ -611,8 +618,7 @@ handle_link (char *buffer)
 	/* Bold text or italics text */
 	if (buffer[0] == '.' && (buffer[1] == 'I' || buffer[1] == 'B'))
 	    for (buffer += 2; *buffer == ' ' || *buffer == '\t'; buffer++);
-	strncpy (old, buffer, sizeof (old) - 1);
-	old[sizeof (old) - 1] = 0;
+	g_strlcpy (old, buffer, sizeof (old));
 	link_flag = 3;
 	break;
     case 3:
@@ -633,12 +639,13 @@ handle_link (char *buffer)
 	if (amp) {
 	    *amp = 0;
 	    amp += 2;
+	    amp_arg = amp;
 	} else {
-	    amp = "";
+	    amp_arg = "";
 	}
 
 	printf_string ("%c%s%c%s%c%s\n", CHAR_LINK_START, old,
-		       CHAR_LINK_POINTER, buffer, CHAR_LINK_END, amp);
+		       CHAR_LINK_POINTER, buffer, CHAR_LINK_END, amp_arg);
 	link_flag = 0;
 	/* Add to the linked list */
 	if (current_link) {
@@ -664,7 +671,7 @@ main (int argc, char **argv)
     FILE *f_man;		/* Manual file */
     FILE *f_tmpl;		/* Template file */
     char buffer[BUFFER_SIZE];	/* Full input line */
-    char *node = NULL;
+    char *lc_node = NULL;
     char *outfile_buffer;	/* Large buffer to keep the output file */
     long cont_start;		/* Start of [Contents] */
     long file_end;		/* Length of the output file */
@@ -747,21 +754,21 @@ main (int argc, char **argv)
     /* Repeat for each input line */
     /* Read a line */
     while (fgets (buffer, BUFFER_SIZE, f_tmpl)) {
-	if (node) {
+	if (lc_node) {
 	    if (*buffer && *buffer != '\n') {
 		cnode->lname = strdup (buffer);
-		node = strchr (cnode->lname, '\n');
-		if (node)
-		    *node = 0;
+		lc_node = strchr (cnode->lname, '\n');
+		if (lc_node)
+		    *lc_node = 0;
 	    }
-	    node = NULL;
+	    lc_node = NULL;
 	} else {
-	    node = strchr (buffer, CHAR_NODE_END);
-	    if (node && (node[1] == '[')) {
-		char *p = strchr (node, ']');
+	    lc_node = strchr (buffer, CHAR_NODE_END);
+	    if (lc_node && (lc_node[1] == '[')) {
+		char *p = strchr (lc_node, ']');
 		if (p) {
-		    if (strncmp (node + 1, "[main]", 6) == 0) {
-			node = 0;
+		    if (strncmp (lc_node + 1, "[main]", 6) == 0) {
+			lc_node = NULL;
 		    } else {
 			if (!cnode) {
 			    cnode = &nodes;
@@ -769,16 +776,16 @@ main (int argc, char **argv)
 			    cnode->next = malloc (sizeof (nodes));
 			    cnode = cnode->next;
 			}
-			cnode->node = strdup (node + 2);
-			cnode->node[p - node - 2] = 0;
+			cnode->node = strdup (lc_node + 2);
+			cnode->node[p - lc_node - 2] = 0;
 			cnode->lname = NULL;
 			cnode->next = NULL;
 			cnode->heading_level = 0;
 		    }
 		} else
-		    node = NULL;
+		    lc_node = NULL;
 	    } else
-		node = NULL;
+		lc_node = NULL;
 	}
 	fputs (buffer, f_out);
     }
@@ -822,12 +829,12 @@ main (int argc, char **argv)
     }
 
     for (cnode = &nodes; cnode && cnode->node;) {
-	char *node = cnode->node;
 	struct node *next = cnode->next;
+	lc_node = cnode->node;
 
-	if (*node)
+	if (*lc_node)
 	    fprintf (f_out, "  %*s\001%s\002%s\003", cnode->heading_level,
-		     "", cnode->lname ? cnode->lname : node, node);
+		     "", cnode->lname ? cnode->lname : lc_node, lc_node);
 	fprintf (f_out, "\n");
 
 	free (cnode->node);
