@@ -1,5 +1,6 @@
 /* User Menu implementation
-   Copyright (C) 1994 Miguel de Icaza, Janne Kukonlehto
+   Copyright (C) 1994, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005,
+   2006, 2007 Free Software Foundation, Inc.
    
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -13,45 +14,50 @@
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
-   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  */
+   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.  */
+
+/** \file user.c
+ *  \brief Source: user menu implementation
+ */
 
 #include <config.h>
-#include <stdio.h>
-#ifdef NEEDS_IO_H
-#    include <io.h>
-#endif
-#include "tty.h"
-#include <string.h>
+
 #include <ctype.h>
 #include <errno.h>
-#include <fcntl.h>
-#include "global.h"
-#include "dialog.h"
-#include "color.h"
+#include <stdlib.h>
+#include <stdio.h>
+#include <string.h>
+
+#include "lib/global.h"
+#include "lib/tty/tty.h"
+#include "lib/skin.h"
+#include "lib/search.h"
+#include "lib/vfs/mc-vfs/vfs.h"
+#include "lib/strutil.h"
+
+#include "src/editor/edit.h"		/* WEdit, BLOCK_FILE */
+#include "src/viewer/mcviewer.h" /* for default_* externs */
+
 #include "dir.h"
 #include "panel.h"
 #include "main.h"
-#include "subshell.h" 		/* for subshell_pty */
 #include "user.h"
 #include "layout.h"
+#include "execute.h"
 #include "setup.h"
-#include "../vfs/vfs.h"
+#include "history.h"
+
 
 /* For the simple listbox manager */
-#include "dlg.h"
+#include "dialog.h"
 #include "widget.h"
 #include "wtools.h"
 
-#include "view.h" /* for default_* externs */
-
-/* "$Id: user.c,v 1.20 2000/10/04 16:34:53 proskin Exp $" */
-
 #define MAX_ENTRIES 16
 #define MAX_ENTRY_LEN 60
 
 static int debug_flag = 0;
 static int debug_error = 0;
-static WEdit *s_editwidget;
 static char *menu = NULL;
 
 /* Formats defined:
@@ -63,9 +69,9 @@ static char *menu = NULL;
    %s  "Selected files"; the tagged files if any, otherwise the current file
    %t  Tagged files
    %u  Tagged files (and they are untagged on return from expand_format)
-   %view Runs the commands and pipes standard output to the view command
-       if %view is directly followed by '{', a list of keywords
-       ascii, hex, nroff, unformated and
+   %view Runs the commands and pipes standard output to the view command.
+       If %view is immediately followed by '{', recognize keywords
+       ascii, hex, nroff and unform
 
    If the format letter is in uppercase, it refers to the other panel.
    
@@ -82,22 +88,22 @@ static char *menu = NULL;
 int check_format_view (const char *p)
 {
     const char *q = p;
-    if (!strncmp (p, "view", 4)){
+    if (!strncmp (p, "view", 4)) {
     	q += 4;
     	if (*q == '{'){
-    	    for (q++;*q && *q != '}';q++){
-    	    	if (!strncmp (q, "ascii", 5)){
-    	    	    default_hex_mode = 0;
+    	    for (q++;*q && *q != '}';q++) {
+    	    	if (!strncmp (q, "ascii", 5)) {
+    	    	    mcview_default_hex_mode = 0;
     	    	    q += 4;
-    	    	} else if (!strncmp (q, "hex", 3)){
-    	    	    default_hex_mode = 1;
+    	    	} else if (!strncmp (q, "hex", 3)) {
+    	    	    mcview_default_hex_mode = 1;
     	    	    q += 2;
-    	    	} else if (!strncmp (q, "nroff", 5)){
-    	    	    default_nroff_flag = 1;
+    	    	} else if (!strncmp (q, "nroff", 5)) {
+    	    	    mcview_default_nroff_flag = 1;
     	    	    q += 4;
-    	    	} else if (!strncmp (q, "unformated", 10)){
-    	    	    default_nroff_flag = 0;
-    	    	    q += 9;
+    	    	} else if (!strncmp (q, "unform", 6)) {
+    	    	    mcview_default_nroff_flag = 0;
+    	    	    q += 5;
     	    	} 
     	    }
     	    if (*q == '}')
@@ -120,7 +126,7 @@ int check_format_var (const char *p, char **v)
 {
     const char *q = p;
     char *var_name;
-    char *value;
+    const char *value;
     const char *dots = 0;
     
     *v = 0;
@@ -133,27 +139,23 @@ int check_format_var (const char *p, char **v)
 	    return 0;
 
 	if (!dots || dots == q+5){
-	    message (1,
+	    message (D_ERROR,
 		     _(" Format error on file Extensions File "),
-		     _(!dots ? N_(" The %%var macro does not have a default ")
-		     :         N_(" The %%var macros does not have a variable ")));
+		     !dots ? _(" The %%var macro has no default ")
+		     :       _(" The %%var macro has no variable "));
 	    return 0;
 	}
 	
 	/* Copy the variable name */
-	var_name = g_malloc (dots - p);
-	strncpy (var_name, p+4, dots-2 - (p+3));
-	var_name [dots-2 - (p+3)] = 0;
+	var_name = g_strndup (p + 4, dots-2 - (p+3));
 
 	value = getenv (var_name);
 	g_free (var_name);
 	if (value){
 	    *v = g_strdup (value);
 	    return q-p;
-	}
-	var_name = g_malloc (q - dots + 1);
-	strncpy (var_name, dots, q - dots + 1);
-	var_name [q-dots] = 0;
+	}	
+	var_name = g_strndup (dots, q - dots);
 	*v = var_name;
 	return q-p;
     }
@@ -175,119 +177,194 @@ strip_ext(char *ss)
     return ss;
 }
 
-char *expand_format (char c, int quote)
+char *
+expand_format (struct WEdit *edit_widget, char c, gboolean do_quote)
 {
-    WPanel *panel;
-    char *(*quote_func)(const char *, int);
-    char *fname;
+    WPanel *panel = NULL;
+    char *(*quote_func) (const char *, int);
+    char *fname = NULL;
+    char *result;
+    char c_lc;
+
+#ifndef USE_INTERNAL_EDIT
+    (void) edit_widget;
+#endif
 
     if (c == '%')
 	return g_strdup ("%");
-    
-    if (islower (c))
-	panel = cpanel;
-    else {
-	if (get_other_type () != view_listing)
-	    return g_strdup ("");
-	panel = other_panel;
-    }
-    if (!panel)
-	panel = cpanel;
 
-    if (quote)
+    if (edit_one_file == NULL) {
+	if (g_ascii_islower ((gchar) c))
+	    panel = current_panel;
+	else {
+	    if (get_other_type () != view_listing)
+		return g_strdup ("");
+	    panel = other_panel;
+	}
+	fname = panel->dir.list[panel->selected].fname;
+    }
+#ifdef USE_INTERNAL_EDIT
+    else
+	fname = str_unconst (edit_get_file_name (edit_widget));
+#endif
+
+    if (do_quote)
 	quote_func = name_quote;
     else
 	quote_func = fake_name_quote;
 
-    c = tolower (c);
-    fname = panel->dir.list [panel->selected].fname;
-    
-    switch (c){
-    case 'f': 
-    case 'p': return (*quote_func) (fname, 0);
-    case 'b': return strip_ext((*quote_func) (fname, 0));
-    case 'x': return (*quote_func) (extension(fname), 0);
-    case 'd': return (*quote_func) (panel->cwd, 0);
-    case 'i': /* indent equal number cursor position in line */
-        if (s_editwidget)
-	        return g_strnfill (s_editwidget->curs_col, ' ');
-        break;
-    case 'y': /* syntax type */
-        if (s_editwidget && s_editwidget->syntax_type)
-	        return g_strdup (s_editwidget->syntax_type);
-        break;
-    case 'e': /* error file name */
-    case 'k': /* block file name */ {
-	char *file = g_strconcat (home_dir, (c == 'e') ? ERROR_FILE : BLOCK_FILE, NULL);
-	fname = (*quote_func) (file, 0);
-	g_free (file);
-	return fname;
-    }
-    case 'm': /* menu file name */
+    c_lc = g_ascii_tolower ((gchar) c);
+
+    switch (c_lc) {
+    case 'f':
+    case 'p':
+	return (*quote_func) (fname, 0);
+    case 'x':
+	return (*quote_func) (extension (fname), 0);
+    case 'd':
+	{
+	    char *cwd;
+	    char *qstr;
+
+	    cwd = g_malloc(MC_MAXPATHLEN + 1);
+
+	    if (panel)
+		g_strlcpy(cwd, panel->cwd, MC_MAXPATHLEN + 1);
+	    else
+		mc_get_current_wd(cwd, MC_MAXPATHLEN + 1);
+
+	    qstr = (*quote_func) (cwd, 0);
+
+	    g_free (cwd);
+
+	    return qstr;
+	}
+    case 'i':			/* indent equal number cursor position in line */
+#ifdef USE_INTERNAL_EDIT
+	if (edit_widget)
+	    return g_strnfill (edit_get_curs_col (edit_widget), ' ');
+#endif
+	break;
+    case 'y':			/* syntax type */
+#ifdef USE_INTERNAL_EDIT
+	if (edit_widget) {
+	    const char *syntax_type = edit_get_syntax_type (edit_widget);
+	    if (syntax_type != NULL)
+		return g_strdup (syntax_type);
+	}
+#endif
+	break;
+    case 'k':			/* block file name */
+    case 'b':			/* block file name / strip extension */  {
+#ifdef USE_INTERNAL_EDIT
+	    if (edit_widget) {
+		char *file = concat_dir_and_file (home_dir, EDIT_BLOCK_FILE);
+		fname = (*quote_func) (file, 0);
+		g_free (file);
+		return fname;
+	    }
+#endif
+	    if (c_lc == 'b')
+		return strip_ext ((*quote_func) (fname, 0));
+	    break;
+	}
+    case 'n':			/* strip extension in editor */
+#ifdef USE_INTERNAL_EDIT
+	if (edit_widget)
+	    return strip_ext ((*quote_func) (fname, 0));
+#endif
+	break;
+    case 'm':			/* menu file name */
 	if (menu)
 	    return (*quote_func) (menu, 0);
 	break;
-    case 's': if (!panel->marked)
-		return (*quote_func) (fname, 0);
+    case 's':
+	if (!panel || !panel->marked)
+	    return (*quote_func) (fname, 0);
 
 	/* Fall through */
 
     case 't':
     case 'u':
-    {
-	int length = 2, i;
-	char *block, *tmp;
-	
-	for (i = 0; i < panel->count; i++)
-	    if (panel->dir.list [i].f.marked)
-		length += strlen (panel->dir.list [i].fname) + 1;
+	{
+	    int length = 2, i;
+	    char *block, *tmp;
 
-	block = g_malloc (length*2+1);
-	*block = 0;
-	for (i = 0; i < panel->count; i++)
-	    if (panel->dir.list [i].f.marked){
-		strcat (block, tmp = (*quote_func) (panel->dir.list [i].fname, 0));
-		g_free (tmp);
-		strcat (block, " ");
-		if (c == 'u')
-		    do_file_mark (panel, i, 0);
-	    }
-	return block;
-    } /* sub case block */
-    } /* switch */
-    return g_strdup ("");
+	    if (!panel)
+		return g_strdup ("");
+
+	    for (i = 0; i < panel->count; i++)
+		if (panel->dir.list[i].f.marked)
+		    length += strlen (panel->dir.list[i].fname) + 1;	/* for space */
+
+	    block = g_malloc (length * 2 + 1);
+	    *block = 0;
+	    for (i = 0; i < panel->count; i++)
+		if (panel->dir.list[i].f.marked) {
+		    strcat (block, tmp =
+			    (*quote_func) (panel->dir.list[i].fname, 0));
+		    g_free (tmp);
+		    strcat (block, " ");
+		    if (c_lc == 'u')
+			do_file_mark (panel, i, 0);
+		}
+	    return block;
+	}			/* sub case block */
+    }				/* switch */
+    result = g_strdup ("% ");
+    result[1] = c;
+    return result;
 }
 
-/* Checks for shell patterns definition */
-char *check_patterns (char *p)
+/*
+ * Check for the "shell_patterns" directive.  If it's found and valid,
+ * interpret it and move the pointer past the directive.  Return the
+ * current pointer.
+ */
+static char *
+check_patterns (char *p)
 {
-    static const char def_name [] = "shell_patterns=";
-    int value;
+    static const char def_name[] = "shell_patterns=";
+    char *p0 = p;
 
-    if (strncmp (p, def_name, sizeof (def_name) - 1) == 0){
-	p += sizeof (def_name) - 1;
-	value = *p++ - '0';
-	if (value == 0 || value == 1)
-	    easy_patterns = value;
-	else
-	    message (1, MSG_ERROR, _(" Invalid shell pattern definition \"%c\". "), value + '0');
-    }
-    while (*p == '\n' || *p == '\t' || *p == ' ') p++;
+    if (strncmp (p, def_name, sizeof (def_name) - 1) != 0)
+	return p0;
+
+    p += sizeof (def_name) - 1;
+    if (*p == '1')
+	easy_patterns = 1;
+    else if (*p == '0')
+	easy_patterns = 0;
+    else
+	return p0;
+
+    /* Skip spaces */
+    p++;
+    while (*p == '\n' || *p == '\t' || *p == ' ')
+	p++;
     return p;
 }
 
 /* Copies a whitespace separated argument from p to arg. Returns the
    point after argument. */
-static char *extract_arg (char *p, char *arg)
+static char *extract_arg (char *p, char *arg, int size)
 {
+    char *np;
+
     while (*p && (*p == ' ' || *p == '\t' || *p == '\n'))
 	p++;
                 /* support quote space .mnu */
-    while (*p && (*p != ' ' || *(p-1) == '\\') && *p != '\t' && *p != '\n')
-	*arg++ = *p++;
+    while (*p && (*p != ' ' || *(p-1) == '\\') && *p != '\t' && *p != '\n') {
+	np = str_get_next_char (p);
+	if (np - p >= size) break;
+	memcpy (arg, p, np - p);
+	arg+= np - p;
+	size-= np - p;
+	p = np;
+    }
     *arg = 0;
     if (!*p || *p == '\n')
-	p --;
+	str_prev_char (&p);
     return p;
 }
 
@@ -296,7 +373,7 @@ static char *extract_arg (char *p, char *arg)
 static int test_type (WPanel *panel, char *arg)
 {
     int result = 0; /* False by default */
-    int st_mode = panel->dir.list [panel->selected].buf.st_mode;
+    int st_mode = panel->dir.list [panel->selected].st.st_mode;
 
     for (;*arg != 0; arg++){
 	switch (*arg){
@@ -340,10 +417,18 @@ static int test_type (WPanel *panel, char *arg)
 
 /* Calculates the truth value of the next condition starting from
    p. Returns the point after condition. */
-static char *test_condition (char *p, int *condition)
+static char *
+test_condition (WEdit *edit_widget, char *p, int *condition)
 {
     WPanel *panel;
     char arg [256];
+    mc_search_type_t search_type;
+
+    if (easy_patterns) {
+	search_type = MC_SEARCH_T_GLOB;
+    } else {
+	search_type = MC_SEARCH_T_REGEX;
+    }
 
     /* Handle one condition */
     for (;*p != '\n' && *p != '&' && *p != '|'; p++){
@@ -351,7 +436,7 @@ static char *test_condition (char *p, int *condition)
 	if ((*p == ' ' && *(p-1) != '\\') || *p == '\t')
 	    continue;
 	if (*p >= 'a')
-	    panel = cpanel;
+	    panel = current_panel;
 	else {
 	    if (get_other_type () == view_listing)
 		panel = other_panel;
@@ -362,34 +447,38 @@ static char *test_condition (char *p, int *condition)
 
 	switch (*p++){
 	case '!':
-	    p = test_condition (p, condition);
+	    p = test_condition (edit_widget, p, condition);
 	    *condition = ! *condition;
-	    p--;
+	    str_prev_char (&p);
 	    break;
 	case 'f': /* file name pattern */
-	    p = extract_arg (p, arg);
-	    *condition = panel && regexp_match (arg, panel->dir.list [panel->selected].fname, match_file);
+	    p = extract_arg (p, arg, sizeof (arg));
+	    *condition = panel && mc_search (arg, panel->dir.list [panel->selected].fname, search_type);
 	    break;
 	case 'y': /* syntax pattern */
-            if (s_editwidget && s_editwidget->syntax_type) {
-	        p = extract_arg (p, arg);
-	        *condition = panel &&
-                    regexp_match (arg, s_editwidget->syntax_type, match_normal);
+#ifdef USE_INTERNAL_EDIT
+	    if (edit_widget) {
+		const char *syntax_type = edit_get_syntax_type (edit_widget);
+		if (syntax_type != NULL) {
+		    p = extract_arg (p, arg, sizeof (arg));
+		    *condition = panel && mc_search (arg, syntax_type, MC_SEARCH_T_NORMAL);
+		}
 	    }
-                break;
+#endif
+            break;
 	case 'd':
-	    p = extract_arg (p, arg);
-	    *condition = panel && regexp_match (arg, panel->cwd, match_file);
+	    p = extract_arg (p, arg, sizeof (arg));
+	    *condition = panel && mc_search (arg, panel->cwd, search_type);
 	    break;
 	case 't':
-	    p = extract_arg (p, arg);
+	    p = extract_arg (p, arg, sizeof (arg));
 	    *condition = panel && test_type (panel, arg);
 	    break;
 	case 'x': /* executable */
 	{
 	    struct stat status;
-	    
-	    p = extract_arg (p, arg);
+
+	    p = extract_arg (p, arg, sizeof (arg));
 	    if (stat (arg, &status) == 0)
 		*condition = is_exe (status.st_mode);
 	    else
@@ -409,63 +498,57 @@ static char *test_condition (char *p, int *condition)
 static void
 debug_out (char *start, char *end, int cond)
 {
-    static char msg [256];
+    static char *msg;
     int len;
 
     if (start == NULL && end == NULL){
-	if (cond == 0){
-	    /* Init */
-	    msg [0] = 0;
-	} else {
-	    /* Show output */
-	    if (!debug_flag)
-		return;
+	/* Show output */
+	if (debug_flag && msg) {
 	    len = strlen (msg);
 	    if (len)
 		msg [len - 1] = 0;
-	    message (0, _(" Debug "), msg);
-	    debug_flag = 0;
+	    message (D_NORMAL, _(" Debug "), "%s", msg);
+
 	}
+	debug_flag = 0;
+	g_free (msg);
+	msg = NULL;
     } else {
+	const char *type;
+	char *p;
+
 	/* Save debug info for later output */
 	if (!debug_flag)
 	    return;
 	/* Save the result of the condition */
 	if (debug_error){
-	    strcat (msg, _(" ERROR: "));
+	    type = _(" ERROR: ");
 	    debug_error = 0;
 	}
 	else if (cond)
-	    strcat (msg, _(" True:  "));
+	    type = _(" True:  ");
 	else
-	    strcat (msg, _(" False: "));
-	/* Copy condition statement */
-	len = strlen (msg);
-	if (end == NULL){
-	    /* Copy one character */
-	    msg [len] = *start;
-	    msg [len + 1] = 0;
-	} else {
-	    /* Copy many characters */
-	    while (start < end){
-		msg [len++] = *start++;
-	    }
-	    msg [len] = 0;
-	}
-	strcat (msg, " \n");
+	    type = _(" False: ");
+	/* This is for debugging, don't need to be super efficient.  */
+	if (end == NULL)
+	    p = g_strdup_printf ("%s%s%c \n", msg ? msg : "", type, *start);
+	else
+	    p = g_strdup_printf ("%s%s%.*s \n", msg ? msg : "", type,
+			(int) (end - start), start);
+	g_free (msg);
+	msg = p;
     }
 }
 
 /* Calculates the truth value of one lineful of conditions. Returns
    the point just before the end of line. */
-static char *test_line (char *p, int *result)
+static char *
+test_line (WEdit *edit_widget, char *p, int *result)
 {
     int condition;
     char operator;
     char *debug_start, *debug_end;
 
-    /* Init debugger */
-    debug_out (NULL, NULL, 0);
     /* Repeat till end of line */
     while (*p && *p != '\n') {
         /* support quote space .mnu */
@@ -486,7 +569,7 @@ static char *test_line (char *p, int *result)
 	condition = 1;	/* True by default */
 
 	debug_start = p;
-	p = test_condition (p, &condition);
+	p = test_condition (edit_widget, p, &condition);
 	debug_end = p;
 	/* Add one debug statement */
 	debug_out (debug_start, debug_end, condition);
@@ -515,50 +598,39 @@ static char *test_line (char *p, int *result)
     debug_out (NULL, NULL, 1);
 
     if (!*p || *p == '\n')
-	p --;
+	str_prev_char (&p);
     return p;
 }
 
 /* FIXME: recode this routine on version 3.0, it could be cleaner */
 static void
-execute_menu_command (char *commands)
+execute_menu_command (WEdit *edit_widget, const char *commands)
 {
     FILE *cmd_file;
     int  cmd_file_fd;
     int  expand_prefix_found = 0;
     char *parameter = 0;
-    int  do_quote = 0;
-    char prompt [80];
+    gboolean do_quote = FALSE;
+    char lc_prompt [80];
     int  col;
     char *file_name;
-#ifdef OS2_NT
-    char *p;
-#endif
+    int run_view = 0;
+
     /* Skip menu entry title line */
     commands = strchr (commands, '\n');
     if (!commands){
 	return;
     }
 
-    if ((file_name = tempnam (NULL, "mcusr")) == 0) {
-	message (1, MSG_ERROR, _(" Can't generate unique filename \n %s "),
-		 unix_error_string (errno));
-	return;
-    }
+    cmd_file_fd = mc_mkstemps (&file_name, "mcusr", SCRIPT_SUFFIX);
 
-#ifdef OS2_NT
-    /* OS/2 and NT requires the command to end in .cmd */
-    p = file_name;
-    file_name = g_strconcat (file_name, ".cmd", NULL);
-    free (p);
-#endif
-    if ((cmd_file_fd = open (file_name, O_RDWR | O_CREAT | O_TRUNC | O_EXCL, 0600)) == -1){
-	message (1, MSG_ERROR, _(" Can't create temporary command file \n %s "),
+    if (cmd_file_fd == -1){
+	message (D_ERROR, MSG_ERROR, _(" Cannot create temporary command file \n %s "),
 		 unix_error_string (errno));
-        free (file_name);
 	return;
     }
     cmd_file = fdopen (cmd_file_fd, "w");
+    fputs ("#! /bin/sh\n", cmd_file);
     commands++;
     
     for (col = 0; *commands; commands++){
@@ -567,6 +639,8 @@ execute_menu_command (char *commands)
 		break;
 	    while (*commands == ' ' || *commands == '\t')
 	        commands++;
+	    if (*commands == 0)
+		break;
 	}
 	col++;
 	if (*commands == '\n')
@@ -575,53 +649,69 @@ execute_menu_command (char *commands)
 	    if (*commands == '}'){
 		char *tmp;
 		*parameter = 0;
-		parameter = input_dialog (_(" Parameter "), prompt, "");
+		parameter = input_dialog (_(" Parameter "), lc_prompt, MC_HISTORY_FM_MENU_EXEC_PARAM, "");
 		if (!parameter || !*parameter){
 		    /* User canceled */
 		    fclose (cmd_file);
 		    unlink (file_name);
-                    free (file_name);
+                    g_free (file_name);
 		    return;
 		}
 		if (do_quote) {
-    		    fputs (tmp = name_quote (parameter, 0), cmd_file);
+		    tmp = name_quote (parameter, 0);
+		    fputs (tmp, cmd_file);
 		    g_free (tmp);
 		} else
 		    fputs (parameter, cmd_file);
 		g_free (parameter);
 		parameter = 0;
 	    } else {
-		if (parameter < &prompt [sizeof (prompt) - 1]) {
+		if (parameter < &lc_prompt [sizeof (lc_prompt) - 1]) {
 		    *parameter++ = *commands;
 		} 
 	    }
 	} else if (expand_prefix_found){
 	    expand_prefix_found = 0;
-	    if (isdigit (*commands)) {
-		do_quote = atoi (commands);
-		while (isdigit (*commands))
+	    if (g_ascii_isdigit ((gchar) *commands)) {
+		do_quote = (atoi (commands) != 0);
+		while (g_ascii_isdigit ((gchar) *commands))
 		    commands++;
 	    }
 	    if (*commands == '{')
-		parameter = prompt;
+		parameter = lc_prompt;
 	    else{
-		char *text = expand_format (*commands, do_quote);
+		char *text = expand_format (edit_widget, *commands, do_quote);
 		fputs (text, cmd_file);
 		g_free (text);
 	    }
 	} else {
 	    if (*commands == '%') {
-		do_quote = 1; /* Default: Quote expanded macro */
-		expand_prefix_found = 1;
+		int i = check_format_view (commands + 1);
+		if (i) {
+		    commands += i;
+		    run_view = 1;
+		} else {
+		    do_quote = TRUE; /* Default: Quote expanded macro */
+		    expand_prefix_found = 1;
+		}
 	    } else
 		fputc (*commands, cmd_file);
 	}
     }
     fclose (cmd_file);
     chmod (file_name, S_IRWXU);
-    execute (file_name);
+    if (run_view) {
+	run_view = 0;
+	mcview_viewer (file_name, NULL, &run_view, 0);
+    } else {
+	/* execute the command indirectly to allow execution even
+	 * on no-exec filesystems. */
+	char *cmd = g_strconcat("/bin/sh ", file_name, (char *) NULL);
+	shell_execute (cmd, EXECUTE_HIDE);
+	g_free(cmd);
+    }
     unlink (file_name);
-    free (file_name);
+    g_free (file_name);
 }
 
 /* 
@@ -644,7 +734,7 @@ menu_file_own(char* path)
     }
     if (verbose)
     {
-	message (0, _(" Warning -- ignoring file "),
+	message (D_NORMAL, _(" Warning -- ignoring file "),
 		    _("File %s is not owned by root or you or is world writable.\n"
 		    "Using it may compromise your security"),
 		path
@@ -654,10 +744,11 @@ menu_file_own(char* path)
 }
 
 /*
-    if edit_widget = pointer then it is file menu from cool edit
-    if edit_widget = NULL then routine is invoke from file menu of mc.
-*/
-void user_menu_cmd (WEdit *edit_widget)
+ * If edit_widget is NULL then we are called from the mc menu,
+ * otherwise we are called from the mcedit menu.
+ */
+void
+user_menu_cmd (struct WEdit *edit_widget)
 {
     char *p;
     char *data, **entries;
@@ -665,34 +756,42 @@ void user_menu_cmd (WEdit *edit_widget)
     int  col, i, accept_entry = 1;
     int  selected, old_patterns;
     Listbox *listbox;
-    s_editwidget = edit_widget;
     
     if (!vfs_current_is_local ()){
-	message (1, _(" Oops... "),
-		 _(" I can't run programs while logged on a non local directory "));
+	message (D_ERROR, MSG_ERROR,
+		 _(" Cannot execute commands on non-local filesystems"));
 	return;
     }
-    
-    menu = g_strdup (edit_widget ? CEDIT_LOCAL_MENU : MC_LOCAL_MENU);
+
+    menu = g_strdup (edit_widget ? EDIT_LOCAL_MENU : MC_LOCAL_MENU);
     if (!exist_file (menu) || !menu_file_own (menu)){
 	g_free (menu);
-        menu = concat_dir_and_file \
-                            (home_dir, edit_widget ? CEDIT_HOME_MENU : MC_HOME_MENU);
+	if (edit_widget)
+	    menu = concat_dir_and_file (home_dir, EDIT_HOME_MENU);
+	else
+	    menu = g_build_filename (home_dir, MC_USERCONF_DIR, MC_USERMENU_FILE, NULL);
+
+
 	if (!exist_file (menu)){
 	    g_free (menu);
-	    menu = concat_dir_and_file \
-                        (mc_home, edit_widget ? CEDIT_GLOBAL_MENU : MC_GLOBAL_MENU);
+	    menu = concat_dir_and_file
+                        (mc_home, edit_widget ? EDIT_GLOBAL_MENU : MC_GLOBAL_MENU);
+	    if (!exist_file (menu)) {
+		g_free (menu);
+		menu = concat_dir_and_file
+			(mc_home_alt, edit_widget ? EDIT_GLOBAL_MENU : MC_GLOBAL_MENU);
+	    }
 	}
     }
 
     if ((data = load_file (menu)) == NULL){
-	message (1, MSG_ERROR, _(" Can't open file %s \n %s "),
+	message (D_ERROR, MSG_ERROR, _(" Cannot open file %s \n %s "),
 		 menu, unix_error_string (errno));
 	g_free (menu);
 	menu = NULL;
 	return;
     }
-    
+
     max_cols = 0;
     selected = 0;
     menu_limit = 0;
@@ -701,20 +800,20 @@ void user_menu_cmd (WEdit *edit_widget)
     /* Parse the menu file */
     old_patterns = easy_patterns;
     p = check_patterns (data);
-    for (menu_lines = col = 0; *p; p++){
+    for (menu_lines = col = 0; *p; str_next_char (&p)){
 	if (menu_lines >= menu_limit){
 	    char ** new_entries;
-	    
-	    menu_limit += MAX_ENTRIES;
-	    new_entries = g_realloc (entries, sizeof (new_entries[0]) * menu_limit);
 
-	    if (new_entries == 0)
+	    menu_limit += MAX_ENTRIES;
+	    new_entries = g_try_realloc (entries, sizeof (new_entries[0]) * menu_limit);
+
+	    if (new_entries == NULL)
 		break;
 
 	    entries = new_entries;
 	    new_entries += menu_limit;
 	    while (--new_entries >= &entries[menu_lines])
-		*new_entries = 0;
+		*new_entries = NULL;
 	}
 	if (col == 0 && !entries [menu_lines]){
 	    if (*p == '#'){
@@ -723,28 +822,28 @@ void user_menu_cmd (WEdit *edit_widget)
 	    } else if (*p == '+'){
 		if (*(p+1) == '='){
 		    /* Combined adding and default */
-		    p = test_line (p, &accept_entry);
+		    p = test_line (edit_widget, p + 1, &accept_entry);
 		    if (selected == 0 && accept_entry)
 			selected = menu_lines;
 		} else {
 		    /* A condition for adding the entry */
-		    p = test_line (p, &accept_entry);
+		    p = test_line (edit_widget, p, &accept_entry);
 		}
 	    } else if (*p == '='){
 		if (*(p+1) == '+'){
 		    /* Combined adding and default */
-		    p = test_line (p, &accept_entry);
+		    p = test_line (edit_widget, p + 1, &accept_entry);
 		    if (selected == 0 && accept_entry)
 			selected = menu_lines;
 		} else {
 		    /* A condition for making the entry default */
 		    i = 1;
-		    p = test_line (p, &i);
+		    p = test_line (edit_widget, p, &i);
 		    if (selected == 0 && i)
 			selected = menu_lines;
 		}
 	    }
-	    else if (*p != ' ' && *p != '\t' && is_printable (*p)) {
+	    else if (*p != ' ' && *p != '\t' && str_isprint (p)) {
 		/* A menu entry title line */
 		if (accept_entry)
 		    entries [menu_lines] = p;
@@ -766,36 +865,33 @@ void user_menu_cmd (WEdit *edit_widget)
 	}
     }
 
-    if (menu_lines == 0) {
-	message (1, MSG_ERROR, _(" No appropriative entries found in %s "), menu);
-    } else {
+    if (menu_lines == 0)
+	message (D_ERROR, MSG_ERROR, _(" No suitable entries found in %s "), menu);
+    else {
+	max_cols = min (max (max_cols, col), MAX_ENTRY_LEN);
 
-    max_cols = min (max (max_cols, col), MAX_ENTRY_LEN);
- 
-    /* Create listbox */
-    listbox = create_listbox_window (max_cols+2, menu_lines, _(" User menu "),
-				     "[Menu File Edit]");
-    /* insert all the items found */
-    for (i = 0; i < menu_lines; i++) {
-	p = entries [i];
-	LISTBOX_APPEND_TEXT (listbox, (unsigned char) p[0],
-			     extract_line (p, p + MAX_ENTRY_LEN), p
-			    );
-    }
-    /* Select the default entry */
-    listbox_select_by_number (listbox->list, selected);
-    
-    selected = run_listbox (listbox);
-    if (selected >= 0)
-	execute_menu_command (entries [selected]);
+	/* Create listbox */
+	listbox = create_listbox_window (menu_lines, max_cols + 2,_(" User menu "),
+					    "[Menu File Edit]");
+	/* insert all the items found */
+	for (i = 0; i < menu_lines; i++) {
+	    p = entries [i];
+	    LISTBOX_APPEND_TEXT (listbox, (unsigned char) p[0],
+				extract_line (p, p + MAX_ENTRY_LEN), p);
+	}
+	/* Select the default entry */
+	listbox_select_by_number (listbox->list, selected);
 
-    do_refresh ();
+	selected = run_listbox (listbox);
+	if (selected >= 0)
+	    execute_menu_command (edit_widget, entries [selected]);
+
+	do_refresh ();
     }
 
     easy_patterns = old_patterns;
     g_free (menu);
     menu = NULL;
-    if (entries)
-	g_free (entries);
+    g_free (entries);
     g_free (data);
 }
